package org.eclipse.jdt.internal.compiler.ast;

package .var.folders._j.vfbv187d0bb8b8w5spzgg8180000gn.T;

import java.util.List;
import org.eclipse.jdt.core.compiler.CategorizedProblem;
import org.eclipse.jdt.core.compiler.CharOperation;
import org.eclipse.jdt.internal.compiler.ASTVisitor;
import org.eclipse.jdt.internal.compiler.ClassFile;
import org.eclipse.jdt.internal.compiler.CompilationResult;
import org.eclipse.jdt.internal.compiler.ast.ASTNode;
import org.eclipse.jdt.internal.compiler.ast.Annotation;
import org.eclipse.jdt.internal.compiler.ast.Argument;
import org.eclipse.jdt.internal.compiler.ast.CompilationUnitDeclaration;
import org.eclipse.jdt.internal.compiler.ast.Javadoc;
import org.eclipse.jdt.internal.compiler.ast.Receiver;
import org.eclipse.jdt.internal.compiler.ast.Statement;
import org.eclipse.jdt.internal.compiler.ast.TypeDeclaration;
import org.eclipse.jdt.internal.compiler.ast.TypeParameter;
import org.eclipse.jdt.internal.compiler.ast.TypeReference;
import org.eclipse.jdt.internal.compiler.codegen.CodeStream;
import org.eclipse.jdt.internal.compiler.flow.FlowInfo;
import org.eclipse.jdt.internal.compiler.impl.ReferenceContext;
import org.eclipse.jdt.internal.compiler.lookup.AnnotationBinding;
import org.eclipse.jdt.internal.compiler.lookup.Binding;
import org.eclipse.jdt.internal.compiler.lookup.BlockScope;
import org.eclipse.jdt.internal.compiler.lookup.ClassScope;
import org.eclipse.jdt.internal.compiler.lookup.LocalVariableBinding;
import org.eclipse.jdt.internal.compiler.lookup.LookupEnvironment;
import org.eclipse.jdt.internal.compiler.lookup.MethodBinding;
import org.eclipse.jdt.internal.compiler.lookup.MethodScope;
import org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding;
import org.eclipse.jdt.internal.compiler.lookup.TypeBinding;
import org.eclipse.jdt.internal.compiler.parser.Parser;
import org.eclipse.jdt.internal.compiler.problem.AbortCompilation;
import org.eclipse.jdt.internal.compiler.problem.AbortCompilationUnit;
import org.eclipse.jdt.internal.compiler.problem.AbortMethod;
import org.eclipse.jdt.internal.compiler.problem.AbortType;
import org.eclipse.jdt.internal.compiler.problem.ProblemReporter;
import org.eclipse.jdt.internal.compiler.problem.ProblemSeverities;
import org.eclipse.jdt.internal.compiler.util.Util;

public abstract class AbstractMethodDeclaration extends ASTNode implements ProblemSeverities, ReferenceContext {
  public MethodScope scope;
  
  public char[] selector;
  
  public int declarationSourceStart;
  
  public int declarationSourceEnd;
  
  public int modifiers;
  
  public int modifiersSourceStart;
  
  public Annotation[] annotations;
  
  public Receiver receiver;
  
  public Argument[] arguments;
  
  public TypeReference[] thrownExceptions;
  
  public Statement[] statements;
  
  public int explicitDeclarations;
  
  public MethodBinding binding;
  
  public boolean ignoreFurtherInvestigation = false;
  
  public Javadoc javadoc;
  
  public int bodyStart;
  
  public int bodyEnd = -1;
  
  public CompilationResult compilationResult;
  
  AbstractMethodDeclaration(CompilationResult compilationResult) {
    this.compilationResult = compilationResult;
  }
  
  public void abort(int abortLevel, CategorizedProblem problem) {
    switch (abortLevel) {
      case 2:
        throw new AbortCompilation(this.compilationResult, problem);
      case 4:
        throw new AbortCompilationUnit(this.compilationResult, problem);
      case 8:
        throw new AbortType(this.compilationResult, problem);
    } 
    throw new AbortMethod(this.compilationResult, problem);
  }
  
  public void createArgumentBindings() {
    createArgumentBindings(this.arguments, this.binding, this.scope);
  }
  
  static void createArgumentBindings(Argument[] arguments, MethodBinding binding, MethodScope scope) {
    boolean useTypeAnnotations = scope.environment().usesNullTypeAnnotations();
    if (arguments != null && binding != null)
      for (int i = 0, length = arguments.length; i < length; i++) {
        Argument argument = arguments[i];
        binding.parameters[i] = argument.createBinding(scope, binding.parameters[i]);
        if (!useTypeAnnotations) {
          long argTypeTagBits = argument.binding.tagBits & 0x180000000000000L;
          if (argTypeTagBits != 0L) {
            if (binding.parameterNonNullness == null) {
              binding.parameterNonNullness = new Boolean[arguments.length];
              binding.tagBits |= 0x1000L;
            } 
            binding.parameterNonNullness[i] = Boolean.valueOf((argTypeTagBits == 72057594037927936L));
          } 
        } 
      }  
  }
  
  public void bindArguments() {
    if (this.arguments != null) {
      if (this.binding == null) {
        for (int j = 0, k = this.arguments.length; j < k; j++)
          this.arguments[j].bind(this.scope, null, true); 
        return;
      } 
      boolean used = !(!this.binding.isAbstract() && !this.binding.isNative());
      AnnotationBinding[][] paramAnnotations = null;
      for (int i = 0, length = this.arguments.length; i < length; i++) {
        Argument argument = this.arguments[i];
        this.binding.parameters[i] = argument.bind(this.scope, this.binding.parameters[i], used);
        if (argument.annotations != null) {
          if (paramAnnotations == null) {
            paramAnnotations = new AnnotationBinding[length][];
            for (int j = 0; j < i; j++)
              paramAnnotations[j] = Binding.NO_ANNOTATIONS; 
          } 
          paramAnnotations[i] = argument.binding.getAnnotations();
        } else if (paramAnnotations != null) {
          paramAnnotations[i] = Binding.NO_ANNOTATIONS;
        } 
      } 
      if (paramAnnotations != null)
        this.binding.setParameterAnnotations(paramAnnotations); 
    } 
  }
  
  public void bindThrownExceptions() {
    if (this.thrownExceptions != null && 
      this.binding != null && 
      this.binding.thrownExceptions != null) {
      int thrownExceptionLength = this.thrownExceptions.length;
      int length = this.binding.thrownExceptions.length;
      if (length == thrownExceptionLength) {
        for (int i = 0; i < length; i++)
          (this.thrownExceptions[i]).resolvedType = (TypeBinding)this.binding.thrownExceptions[i]; 
      } else {
        int bindingIndex = 0;
        for (int i = 0; i < thrownExceptionLength && bindingIndex < length; i++) {
          TypeReference thrownException = this.thrownExceptions[i];
          ReferenceBinding thrownExceptionBinding = this.binding.thrownExceptions[bindingIndex];
          char[][] bindingCompoundName = thrownExceptionBinding.compoundName;
          if (bindingCompoundName != null)
            if (thrownException instanceof org.eclipse.jdt.internal.compiler.ast.SingleTypeReference) {
              int lengthName = bindingCompoundName.length;
              char[] thrownExceptionTypeName = thrownException.getTypeName()[0];
              if (CharOperation.equals(thrownExceptionTypeName, bindingCompoundName[lengthName - 1])) {
                thrownException.resolvedType = (TypeBinding)thrownExceptionBinding;
                bindingIndex++;
              } 
            } else if (CharOperation.equals(thrownException.getTypeName(), bindingCompoundName)) {
              thrownException.resolvedType = (TypeBinding)thrownExceptionBinding;
              bindingIndex++;
            }  
        } 
      } 
    } 
  }
  
  static void analyseArguments(LookupEnvironment environment, FlowInfo flowInfo, Argument[] methodArguments, MethodBinding methodBinding) {
    if (methodArguments != null) {
      boolean usesNullTypeAnnotations = environment.usesNullTypeAnnotations();
      int length = Math.min(methodBinding.parameters.length, methodArguments.length);
      for (int i = 0; i < length; i++) {
        if (usesNullTypeAnnotations) {
          long tagBits = (methodBinding.parameters[i]).tagBits & 0x180000000000000L;
          if (tagBits == 72057594037927936L) {
            flowInfo.markAsDefinitelyNonNull((methodArguments[i]).binding);
          } else if (tagBits == 36028797018963968L) {
            flowInfo.markPotentiallyNullBit((methodArguments[i]).binding);
          } else if (methodBinding.parameters[i].isFreeTypeVariable()) {
            flowInfo.markNullStatus((methodArguments[i]).binding, 48);
          } 
        } else if (methodBinding.parameterNonNullness != null) {
          Boolean nonNullNess = methodBinding.parameterNonNullness[i];
          if (nonNullNess != null)
            if (nonNullNess.booleanValue()) {
              flowInfo.markAsDefinitelyNonNull((methodArguments[i]).binding);
            } else {
              flowInfo.markPotentiallyNullBit((methodArguments[i]).binding);
            }  
        } 
        flowInfo.markAsDefinitelyAssigned((methodArguments[i]).binding);
      } 
    } 
  }
  
  public CompilationResult compilationResult() {
    return this.compilationResult;
  }
  
  public void generateCode(ClassScope classScope, ClassFile classFile) {
    classFile.codeStream.wideMode = false;
    if (this.ignoreFurtherInvestigation) {
      if (this.binding == null)
        return; 
      CategorizedProblem[] problems = 
        (this.scope.referenceCompilationUnit()).compilationResult.getProblems();
      int problemsLength;
      CategorizedProblem[] problemsCopy = new CategorizedProblem[problemsLength = problems.length];
      System.arraycopy(problems, 0, problemsCopy, 0, problemsLength);
      classFile.addProblemMethod(this, this.binding, problemsCopy);
      return;
    } 
    int problemResetPC = 0;
    CompilationResult unitResult = null;
    int problemCount = 0;
    if (classScope != null) {
      TypeDeclaration referenceContext = classScope.referenceContext;
      if (referenceContext != null) {
        unitResult = referenceContext.compilationResult();
        problemCount = unitResult.problemCount;
      } 
    } 
    boolean restart = false;
    boolean abort = false;
    do {
      try {
        problemResetPC = classFile.contentsOffset;
        generateCode(classFile);
        restart = false;
      } catch (AbortMethod e) {
        if (e.compilationResult == CodeStream.RESTART_IN_WIDE_MODE) {
          classFile.contentsOffset = problemResetPC;
          classFile.methodCount--;
          classFile.codeStream.resetInWideMode();
          if (unitResult != null)
            unitResult.problemCount = problemCount; 
          restart = true;
        } else if (e.compilationResult == CodeStream.RESTART_CODE_GEN_FOR_UNUSED_LOCALS_MODE) {
          classFile.contentsOffset = problemResetPC;
          classFile.methodCount--;
          classFile.codeStream.resetForCodeGenUnusedLocals();
          if (unitResult != null)
            unitResult.problemCount = problemCount; 
          restart = true;
        } else {
          restart = false;
          abort = true;
        } 
      } 
    } while (restart);
    if (abort) {
      CategorizedProblem[] problems = 
        (this.scope.referenceCompilationUnit()).compilationResult.getAllProblems();
      int problemsLength;
      CategorizedProblem[] problemsCopy = new CategorizedProblem[problemsLength = problems.length];
      System.arraycopy(problems, 0, problemsCopy, 0, problemsLength);
      classFile.addProblemMethod(this, this.binding, problemsCopy, problemResetPC);
    } 
  }
  
  public void generateCode(ClassFile classFile) {
    classFile.generateMethodInfoHeader(this.binding);
    int methodAttributeOffset = classFile.contentsOffset;
    int attributeNumber = classFile.generateMethodInfoAttributes(this.binding);
    if (!this.binding.isNative() && !this.binding.isAbstract()) {
      int codeAttributeOffset = classFile.contentsOffset;
      classFile.generateCodeAttributeHeader();
      CodeStream codeStream = classFile.codeStream;
      codeStream.reset(this, classFile);
      this.scope.computeLocalVariablePositions(this.binding.isStatic() ? 0 : 1, codeStream);
      if (this.arguments != null)
        for (int i = 0, max = this.arguments.length; i < max; i++) {
          LocalVariableBinding argBinding;
          codeStream.addVisibleLocalVariable(argBinding = (this.arguments[i]).binding);
          argBinding.recordInitializationStartPC(0);
        }  
      if (this.statements != null)
        for (int i = 0, max = this.statements.length; i < max; i++)
          this.statements[i].generateCode((BlockScope)this.scope, codeStream);  
      if (this.ignoreFurtherInvestigation)
        throw new AbortMethod((this.scope.referenceCompilationUnit()).compilationResult, null); 
      if ((this.bits & 0x40) != 0)
        codeStream.return_(); 
      codeStream.exitUserScope((BlockScope)this.scope);
      codeStream.recordPositionsFrom(0, this.declarationSourceEnd);
      try {
        classFile.completeCodeAttribute(codeAttributeOffset);
      } catch (NegativeArraySizeException negativeArraySizeException) {
        throw new AbortMethod((this.scope.referenceCompilationUnit()).compilationResult, null);
      } 
      attributeNumber++;
    } else {
      checkArgumentsSize();
    } 
    classFile.completeMethodInfo(this.binding, methodAttributeOffset, attributeNumber);
  }
  
  public void getAllAnnotationContexts(int targetType, List allAnnotationContexts) {}
  
  private void checkArgumentsSize() {
    TypeBinding[] parameters = this.binding.parameters;
    int size = 1;
    for (int i = 0, max = parameters.length; i < max; i++) {
      switch ((parameters[i]).id) {
        case 7:
        case 8:
          size += 2;
          break;
        default:
          size++;
          break;
      } 
      if (size > 255)
        this.scope.problemReporter().noMoreAvailableSpaceForArgument(this.scope.locals[i], (ASTNode)(this.scope.locals[i]).declaration); 
    } 
  }
  
  public CompilationUnitDeclaration getCompilationUnitDeclaration() {
    if (this.scope != null)
      return (this.scope.compilationUnitScope()).referenceContext; 
    return null;
  }
  
  public boolean hasErrors() {
    return this.ignoreFurtherInvestigation;
  }
  
  public boolean isAbstract() {
    if (this.binding != null)
      return this.binding.isAbstract(); 
    return ((this.modifiers & 0x400) != 0);
  }
  
  public boolean isAnnotationMethod() {
    return false;
  }
  
  public boolean isClinit() {
    return false;
  }
  
  public boolean isConstructor() {
    return false;
  }
  
  public boolean isDefaultConstructor() {
    return false;
  }
  
  public boolean isDefaultMethod() {
    return false;
  }
  
  public boolean isInitializationMethod() {
    return false;
  }
  
  public boolean isMethod() {
    return false;
  }
  
  public boolean isNative() {
    if (this.binding != null)
      return this.binding.isNative(); 
    return ((this.modifiers & 0x100) != 0);
  }
  
  public boolean isStatic() {
    if (this.binding != null)
      return this.binding.isStatic(); 
    return ((this.modifiers & 0x8) != 0);
  }
  
  public abstract void parseStatements(Parser paramParser, CompilationUnitDeclaration paramCompilationUnitDeclaration);
  
  public StringBuffer print(int tab, StringBuffer output) {
    if (this.javadoc != null)
      this.javadoc.print(tab, output); 
    printIndent(tab, output);
    printModifiers(this.modifiers, output);
    if (this.annotations != null) {
      printAnnotations(this.annotations, output);
      output.append(' ');
    } 
    TypeParameter[] typeParams = typeParameters();
    if (typeParams != null) {
      output.append('<');
      int max = typeParams.length - 1;
      for (int j = 0; j < max; j++) {
        typeParams[j].print(0, output);
        output.append(", ");
      } 
      typeParams[max].print(0, output);
      output.append('>');
    } 
    printReturnType(0, output).append(this.selector).append('(');
    if (this.receiver != null)
      this.receiver.print(0, output); 
    if (this.arguments != null)
      for (int i = 0; i < this.arguments.length; i++) {
        if (i > 0 || this.receiver != null)
          output.append(", "); 
        this.arguments[i].print(0, output);
      }  
    output.append(')');
    if (this.thrownExceptions != null) {
      output.append(" throws ");
      for (int i = 0; i < this.thrownExceptions.length; i++) {
        if (i > 0)
          output.append(", "); 
        this.thrownExceptions[i].print(0, output);
      } 
    } 
    printBody(tab + 1, output);
    return output;
  }
  
  public StringBuffer printBody(int indent, StringBuffer output) {
    if (isAbstract() || (this.modifiers & 0x1000000) != 0)
      return output.append(';'); 
    output.append(" {");
    if (this.statements != null)
      for (int i = 0; i < this.statements.length; i++) {
        output.append('\n');
        this.statements[i].printStatement(indent, output);
      }  
    output.append('\n');
    printIndent((indent == 0) ? 0 : (indent - 1), output).append('}');
    return output;
  }
  
  public StringBuffer printReturnType(int indent, StringBuffer output) {
    return output;
  }
  
  public void resolve(ClassScope upperScope) {
    if (this.binding == null)
      this.ignoreFurtherInvestigation = true; 
    try {
      bindArguments();
      resolveReceiver();
      bindThrownExceptions();
      resolveJavadoc();
      resolveAnnotations((BlockScope)this.scope, this.annotations, (Binding)this.binding, isConstructor());
      long sourceLevel = (this.scope.compilerOptions()).sourceLevel;
      if (sourceLevel < 3407872L)
        validateNullAnnotations(this.scope.environment().usesNullTypeAnnotations()); 
      resolveStatements();
      if (this.binding != null && (
        this.binding.getAnnotationTagBits() & 0x400000000000L) == 0L && (
        this.binding.modifiers & 0x100000) != 0 && 
        sourceLevel >= 3211264L)
        this.scope.problemReporter().missingDeprecatedAnnotationForMethod(this); 
    } catch (AbortMethod abortMethod) {
      this.ignoreFurtherInvestigation = true;
    } 
  }
  
  public void resolveReceiver() {
    if (this.receiver == null)
      return; 
    if (this.receiver.modifiers != 0)
      this.scope.problemReporter().illegalModifiers(this.receiver.declarationSourceStart, this.receiver.declarationSourceEnd); 
    TypeBinding resolvedReceiverType = this.receiver.type.resolvedType;
    if (this.binding == null || resolvedReceiverType == null || !resolvedReceiverType.isValidBinding())
      return; 
    ReferenceBinding declaringClass = this.binding.declaringClass;
    if (isStatic() || declaringClass.isAnonymousType()) {
      this.scope.problemReporter().disallowedThisParameter(this.receiver);
      return;
    } 
    ReferenceBinding enclosingReceiver = this.scope.enclosingReceiverType();
    if (isConstructor()) {
      if (declaringClass.isStatic() || (
        declaringClass.tagBits & 0x18L) == 0L) {
        this.scope.problemReporter().disallowedThisParameter(this.receiver);
        return;
      } 
      enclosingReceiver = enclosingReceiver.enclosingType();
    } 
    char[][] tokens = (this.receiver.qualifyingName == null) ? null : this.receiver.qualifyingName.getName();
    if (isConstructor()) {
      if (tokens == null || tokens.length > 1 || !CharOperation.equals(enclosingReceiver.sourceName(), tokens[0])) {
        this.scope.problemReporter().illegalQualifierForExplicitThis(this.receiver, (TypeBinding)enclosingReceiver);
        this.receiver.qualifyingName = null;
      } 
    } else if (tokens != null && tokens.length > 0) {
      this.scope.problemReporter().illegalQualifierForExplicitThis2(this.receiver);
      this.receiver.qualifyingName = null;
    } 
    if (TypeBinding.notEquals((TypeBinding)enclosingReceiver, resolvedReceiverType))
      this.scope.problemReporter().illegalTypeForExplicitThis(this.receiver, (TypeBinding)enclosingReceiver); 
    if (this.receiver.type.hasNullTypeAnnotation(TypeReference.AnnotationPosition.ANY))
      this.scope.problemReporter().nullAnnotationUnsupportedLocation(this.receiver.type); 
  }
  
  public void resolveJavadoc() {
    if (this.binding == null)
      return; 
    if (this.javadoc != null) {
      this.javadoc.resolve(this.scope);
      return;
    } 
    if (this.binding.declaringClass != null && !this.binding.declaringClass.isLocalType()) {
      int javadocVisibility = this.binding.modifiers & 0x7;
      ClassScope classScope = this.scope.classScope();
      ProblemReporter reporter = this.scope.problemReporter();
      int severity = reporter.computeSeverity(-1610612250);
      if (severity != 256) {
        if (classScope != null)
          javadocVisibility = Util.computeOuterMostVisibility(classScope.referenceType(), javadocVisibility); 
        int javadocModifiers = this.binding.modifiers & 0xFFFFFFF8 | javadocVisibility;
        reporter.javadocMissing(this.sourceStart, this.sourceEnd, severity, javadocModifiers);
      } 
    } 
  }
  
  public void resolveStatements() {
    if (this.statements != null) {
      for (int i = 0, length = this.statements.length; i < length; i++)
        this.statements[i].resolve((BlockScope)this.scope); 
    } else if ((this.bits & 0x8) != 0 && (
      !isConstructor() || this.arguments != null)) {
      this.scope.problemReporter().undocumentedEmptyBlock(this.bodyStart - 1, this.bodyEnd + 1);
    } 
  }
  
  public void tagAsHavingErrors() {
    this.ignoreFurtherInvestigation = true;
  }
  
  public void tagAsHavingIgnoredMandatoryErrors(int problemId) {}
  
  public void traverse(ASTVisitor visitor, ClassScope classScope) {}
  
  public TypeParameter[] typeParameters() {
    return null;
  }
  
  void validateNullAnnotations(boolean useTypeAnnotations) {
    if (this.binding == null)
      return; 
    if (!useTypeAnnotations) {
      if (this.binding.parameterNonNullness != null) {
        int length = this.binding.parameters.length;
        for (int i = 0; i < length; i++) {
          if (this.binding.parameterNonNullness[i] != null) {
            long nullAnnotationTagBit = this.binding.parameterNonNullness[i].booleanValue() ? 
              72057594037927936L : 36028797018963968L;
            if (!this.scope.validateNullAnnotation(nullAnnotationTagBit, (this.arguments[i]).type, (this.arguments[i]).annotations))
              this.binding.parameterNonNullness[i] = null; 
          } 
        } 
      } 
    } else {
      int length = this.binding.parameters.length;
      for (int i = 0; i < length; i++)
        this.scope.validateNullAnnotation((this.binding.parameters[i]).tagBits, (this.arguments[i]).type, (this.arguments[i]).annotations); 
    } 
  }
}
